/* queue.c - query cache file
 *
 * Copyright (C) 2006-2007 Abilis
 * Copyright (C) 2006-2007 Antek srl
 *
 * Author: Elisa Manara <e@entropika.net>
 * Last update: Fri Dec  1 17:04:49 CET 2006
 *
 * This file is part of Abilis Collector for *nix.
 *
 * Abilis Collector is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * Abilis Collector is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 *                                   
 * You should have received a copy of the GNU General Public License along                        
 * with Abilis Collector; see the file COPYING.  If not, write to the Free                        
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. 
 */  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "define.h"

int mysqlPut(struct CfgData *config, char *query);

static int queueLength(void);
static int queueReset(struct CfgData *config);

/* set queue length and file  position  indicator */
int queueInit(struct CfgData *config)
{
    int lines;

    lines=queueLength();
    if(lines == -1) return -1;
    config->queue.length=lines;
    config->queue.position=SEEK_SET;

    printDebug("Query queue initialized.\n");
    return 0;
}

/* empty queue file and reset struct values */
int queueReset(struct CfgData *config)
{
    /* TODO: check if pos.ind. is at EOF? */
    FILE *q;

    if((q=fopen(QUEUE_FILE, "w")) == NULL) {
	printError("[QUEUE] Cannot reset queue file: %s\n", 
	    strerror(errno));
	return -1;
    }

    fclose(q);
    config->queue.length=0;
    config->queue.position=SEEK_SET;

    printDebug("[QUEUE] Query queue reset.\n");
    return 0;
}

/* return number of queries in queue file, 
 * or -1 in case of error
 */
int queueLength(void)
{
    FILE *q;
    char line[MYSQL_MAX_QLEN];
    int lines=0;

    if((q=fopen(QUEUE_FILE, "r")) == NULL) {
	/* queue file does not exist, nothing to do yet */
	if(errno == ENOENT) return 0;
	/* error here */
	printError("Cannot open queue file: %s\n", strerror(errno));
	return -1;
    }
    while(fgets(line, MYSQL_MAX_QLEN, q)) 
	lines++;
    
    fclose(q);
    return lines;
}

/* add a query line to cache file;
 * return values:
 * -1 in case of errors
 *  0 otherwise
 */
int queueAdd(struct CfgData *config, struct CfgClient *client, char *query)
{
    FILE *q;
    int blen;

    if((q=fopen(QUEUE_FILE, "a")) == NULL) {
	printError("[%s] Cannot open queue file for append: %s\n",
	    client->id, strerror(errno));
	return -1;
    }

    strncat(query, "\n", 1); 
    blen=fwrite(query, strlen(query), 1, q);
    if(blen != 1) {
	printError("[%s] Cannot append to queue file: %s\n",
	    client->id, strerror(errno));
	fclose(q);
	return -1;
    }
    printDebug("[%s] Query saved on queue file.\n", client->id);
    fflush(q);
    fclose(q);
    /* update queue length */
    config->queue.length++;
    return 0;
}

/* queue process abstract:
 * - open queue file
 * - if the first available char is EOF, return (queue empty)
 * - read and process QUEUE_MAX_PROCESS lines
 * - in case of lost connection, retry later
 * - update the file position indicator
 * - if EOF is reached (all queries processed), clear queue file
 */
int queueProcess(struct CfgData *config, time_t now)
{
    FILE *q;
    char query[MYSQL_MAX_QLEN];
    int i, retval;

    /* check time */

    if(config->queue.nextCheck > now) return 0;

    config->queue.nextCheck = now + QUEUE_INTERVAL;
    printDebug("[QUEUE] Processing queries from queue.\n");

    if((q=fopen(QUEUE_FILE, "r")) == NULL) {
	/* queue file does not exist, nothing to do yet */
	if(errno == ENOENT) return 0;
	/* error here */
	printError("Cannot process queue file: %s\n", strerror(errno));
	return -1;
    }

    if((getc(q)) == EOF) {
	fclose(q);
	printDebug("[QUEUE] Queue is empty.\n");
	return 0;
    }

    /* position the indicator */
    fseek(q, config->queue.position, SEEK_SET);

    for(i = 0; i <= QUEUE_MAX_PROCESS; i++) {
	/* read a query line */
	if((fgets(query, MYSQL_MAX_QLEN, q)) == NULL) {
	    fclose(q);
	    return queueReset(config);
	}
	/* try to query MySQL */
	retval=mysqlPut(config, query);
	/* retval ==  0 => ok, continue
	 * retval == -1 => skip malformed query
	 * retval ==  1 => lost connection, retry later
	 * retval ==  2 => duplicate query
	 */
	switch(retval) {
	case 0:
	    printDebug("[QUEUE] Query successfully processed.\n");
	    config->queue.inserted++;
	    break;
	case 1:
	    printDebug("[QUEUE] Lost connection: retry later.\n");
	    break;
	case 2:
	    printDebug("[QUEUE] Duplicate query skipped.\n");
	    config->queue.duplicate++;
	    break;
	default:
	    printError("[QUEUE] Malformed query or unknown error. Skipped.\n");
	    config->queue.skipped++;
	    break;
	}
	/* must restart from here next time; suspend processing now */
	if(retval == 1) break;
	/* update number of processed queries */
	config->queue.processed++;
	/* update file position indicator */
	config->queue.position=ftell(q);
    }

    fclose(q);
    return 0;
}

/* 
 * vim: softtabstop=4 shiftwidth=4 
 */
